

import curses
from .constants import INT_INFINITY
from .drawtarget import DrawTarget

from .strwidth import charwidth


class CursedPad(DrawTarget):
	
	def __init__(self, scr, colours, width, height):
		self.scr = scr
		self.colours = colours
		self.width = width
		self.height = height
		self.pad = curses.newpad(max(height, 1), max(width, 1))
		self.clear()
	
	def resize(self, width, height):
		self.width = width
		self.height = height
		self.pad.resize(max(height, 1), max(width, 1))
		self.clear()
	
	
	def clear(self):
		self.pad.clear()
	
	def write(self, x, y, text, style=None):
		if y >= self.height:
			return
		for char in text:
			w = charwidth(char)
			if x + w > self.width:
				break
			self.set_char(x, y, char, style)
			#if w == 2:
				#self.delete(x + 1, y)
			x += w
			
	
	def set_char(self, x, y, char, style=None):
		try:
			self.pad.addstr(y, x, char, self.colours.attrs(style))
		except curses.error:
			# ncurses has a weird problem:
			# it always raises an error when drawing to the last character in the window
			# it draws first and then raises the error
			# therefore to draw in the last place of the window the last character needs to be ingored
			# other solutions might be possible, but are more hacky
			pass
	
	def draw_pad(self, src, dest_x=0, dest_y=0, width=INT_INFINITY, height=INT_INFINITY, src_x=0, src_y=0):
		raise NotImplementedError()






import curses

from .constants import INT_INFINITY
from .basescreen import BaseScreen
from .textstyle import TextStyle as Style
from .strwidth import charwidth
from .cursedpad import CursedPad




class Colours:
	
	def __init__(self):
		
		self.colours = min(curses.COLORS, 16)
		self.pairs = self.colours*self.colours

		curses.use_default_colors()
		for i in range(0, self.pairs):
			curses.init_pair(i, i%self.colours, i//self.colours)
	
	def get(self, fg=0, bg=0):
		if fg is None:
			fg = 0
		if bg is None:
			bg = 0
		if self.colours == 16:
			return curses.color_pair(fg + bg*self.colours)
		elif self.colours == 8:
			dfg = fg % 8
			dbg = bg % 8
			if bg == 8:
				dbg = 7
			if fg == 8:
				dfg = 7
			colour = curses.color_pair(dfg + dbg*self.colours)
			if fg >= 8 and bg < 8:
				colour |= curses.A_BOLD
			elif fg < 8 and bg >= 8:
				colour |= curses.A_DIM
			return colour
		else:
			return curses.color_pair(0)
	
	def attrs(self, style):
		if style is None:
			style = Style.default
		attributes = self.get(style.fg, style.bg);
		if style.attr[Style.BOLD]:
			attributes |= curses.A_BOLD
		if style.attr[Style.UNDERSCORE]:
			attributes |= curses.A_UNDERLINE
		if style.attr[Style.REVERSE]:
			attributes |= curses.A_REVERSE
		return attributes


class CursedScreen(BaseScreen):
	
	def __init__(self, **_kwargs):
		pass
	
	def initialize_terminal(self):
		# Initialize curses
		self.screen = curses.initscr()

		# Turn off echoing of keys, and enter cbreak mode,
		# where no buffering is performed on keyboard input
		curses.noecho()
		curses.cbreak()

		# In keypad mode, escape sequences for special keys
		# (like the cursor keys) will be interpreted and
		# a special value like curses.KEY_LEFT will be returned
		self.screen.keypad(1)
		
		try:
			# make the cursor invisible
			# store the previous cursor location
			self.cursor_visibility = curses.curs_set(0)
		except:
			# having the cursor displayed is annoying somewhat annoying,
			# but it's not terrible so safe to ignore
			pass

		# Start color, too.  Harmless if the terminal doesn't have
		# color; user can test with has_color() later on.  The try/catch
		# works around a minor bit of over-conscientiousness in the curses
		# module -- the error return from C start_color() is ignorable.
		try:
			curses.start_color()
		except:
			pass
		self.colours = Colours()
		
		self.width = 0
		self.height = 0
		#self.blink_bright_background = blink_bright_background # use the blink attribute for bright backgrounds
		#self.always_reset = always_reset or blink_bright_background # always reset if the style is different than the previous one
		self.update_size()
		self.screen.refresh()
	
	def finalize_terminal(self):
		self.screen.keypad(0)
		curses.echo()
		curses.nocbreak()
		curses.endwin()
		
	def create_pad(self, width, height):
		return CursedPad(self.screen, self.colours, width, height)
	
	
	def update_size(self):
		curses.update_lines_cols()
		(y, x) = self.screen.getmaxyx()
		self.width = x
		self.height = y
	
	def _addstr(self, y, x, string, *args):
		if not self.screen:
			return
		try:
			self.screen.addstr(y, x, string, *args)
		except curses.error:
			# ncurses has a weird problem:
			# it always raises an error when drawing to the last character in the window
			# it draws first and then raises the error
			# therefore to draw in the last place of the window the last character needs to be ingored
			# other solutions might be possible, but are more hacky
			pass
		
	def write(self, x, y, text, style=None):
		self._addstr(y, x, text, self.colours.attrs(style))
	
	def clear(self):
		self.screen.clear();
	
	def reset(self):
		curses.endwin()
		self.screen.refresh()
		self.update_size()
	
	def draw_pad(self, pad, scr_x=0, scr_y=0, width=INT_INFINITY, height=INT_INFINITY, pad_x=0, pad_y=0):
		pad.pad.noutrefresh(pad_y, pad_x, scr_y, scr_x, scr_y + height-1, scr_x + width-1)
	
	def update(self):
		self.screen.refresh()

Screen = CursedScreen





from . import Widget

class Adapt(Widget):
	
	def __init__(self, children, selected=0):
		self.layouts = []
		self.replacements = {}
		for child in self.children:
			if isinstance(child.widget, Layout):
				self.layouts.append(child)
			elif child.key is not None:
				self.replacements[child.key] = child
			else:
				raise ValueError("children of Adapt element should be Layout elements or have a key")
	
	def resize(self, target):
		for child in self.layouts:
			if child.widget.match_size(target.width, target.height
	
	def update(self, target, force):
		if self.is_changed():
			force = True
			self.unchange()
		return self.children[self.selected].update(force) or force
	
	
	@classmethod
	def from_xml(cls, children, attr, text):
		return cls(children, attr.get("selected", int(attr.get("selected-val", 0))))





from .window import Window
from .layout import Layout
from .drawtarget import DrawTarget
from .constants import INT_INFINITY

class BaseScreen(DrawTarget):
	
	def create_window(self, target, x=0, y=0, width=None, height=None):
		return Window(target, x, y, width, height)

	def create_pad(self, width, height):
		raise NotImplementedError()

	def initialize_terminal(self):
		raise NotImplementedError()

	def finalize_terminal(self):
		raise NotImplementedError()

	def update_size(self):
		raise NotImplementedError()
		
	def write(self, x, y, text, style=None):
		raise NotImplementedError()
	
	def clear(self):
		raise NotImplementedError()
	
	def reset(self):
		self.update_size()
		self.clear()
	
	def draw_pad(self, pad, scr_x=0, scr_y=0, width=INT_INFINITY, height=INT_INFINITY, pad_x=0, pad_y=0):
		raise NotImplementedError()
	
	def update(self):
		raise NotImplementedError()



import setuptools

with open("README.md", "r") as fh:
	long_description = fh.read()

setuptools.setup(
	name="ratuil",
	version="0.2.0",
	author="troido",
	author_email="troido@protonmail.com",
	description="A terminal UI library for games",
	long_description=long_description,
	long_description_content_type="text/markdown",
	url="https://github.com/jmdejong/ratuil",
	packages=setuptools.find_packages(),
	classifiers=[
		"Programming Language :: Python :: 3",
		"License :: OSI Approved :: MIT License",
		"Operating System :: OS Independent"
	],
	python_requires='>=3.6'
)

